@@ -43,17 +43,19 @@
 struct mdns_svc {
         char *name;
         enum rr_type type;
-        mdns_callback callback;
+        union
+        {
+                mdns_announce_callback announce_callback;
+                //mdns_listen_callback listen_callback;  // currently unused
+        };
         void *p_cookie;
         struct mdns_svc *next;
 };
 
 struct mdns_conn {
         sock_t sock;
-        // Since windows doesn't use a regular sockaddr struct, we have to keep
-        // track of the protocol family
-        unsigned int family;
-        multicast_if if_addr;
+        multicast_if if_addr;  // NB: In Windows this is the interface index
+        struct mdns_ip mdns_ip;  // IP address and family of the interface
 };
 
 struct mdns_ctx {
@@ -73,7 +75,7 @@ extern void rr_print(const struct rr_entry *);
 extern void rr_free(struct rr_entry *);
 
 #ifndef _WIN32
-#if HAVE_IFADDRS
+#if HAVE_GETIFADDRS
 
 static bool
 mdns_is_interface_valuable(struct ifaddrs* ifa)
@@ -89,9 +91,10 @@ mdns_is_interface_valuable(struct ifaddrs* ifa)
 }
 
 static int
-mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
+mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size_t* p_nb_intf, int ai_family)
 {
         struct ifaddrs *ifs;
+        struct mdns_ip *mdns_ips;
         struct ifaddrs *c;
         size_t nb_if;
         multicast_if* intfs;
@@ -116,12 +119,27 @@ mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
                 freeifaddrs(ifs);
                 return (MDNS_ERROR);
         }
+        *pp_mdns_ips = mdns_ips = malloc(sizeof(*mdns_ips) * nb_if);
+        if (mdns_ips == NULL) {
+                free(intfs);
+                freeifaddrs(ifs);
+                return (MDNS_ERROR);
+        }
         for (c = ifs; c != NULL; c = c->ifa_next) {
                 if (c->ifa_addr == NULL ||
                     c->ifa_addr->sa_family != ai_family ||
                     !mdns_is_interface_valuable(c))
                         continue;
                 memcpy(intfs, c->ifa_addr, sizeof(*intfs));
+                if (ai_family == AF_INET) {
+                        struct sockaddr_in *addr_in = (struct sockaddr_in *)c->ifa_addr;
+                        memcpy(&(*mdns_ips).ipv4.addr, &addr_in->sin_addr, sizeof(struct in_addr));
+                }
+                else {
+                        struct sockaddr_in6 *addr_in = (struct sockaddr_in6 *)c->ifa_addr;
+                        memcpy(&(*mdns_ips).ipv6.addr, &addr_in->sin6_addr, sizeof(struct in6_addr));
+                }
+                mdns_ips++;
                 intfs++;
         }
         freeifaddrs(ifs);
@@ -130,16 +148,24 @@ mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
 }
 #else
 static size_t
-mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
+mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size_t* p_nb_intf, int ai_family)
 {
-        *pp_intfs = malloc(sizeof(**pp_intfs));
-        if (*pp_intfs == NULL)
+        multicast_if *intfs;
+        struct mdns_ip *mdns_ips;
+        *pp_intfs = intfs = malloc(sizeof(*intfs));
+        if (intfs == NULL)
                 return (MDNS_ERROR);
-        **pp_intfs = NULL;
+        memset(intfs, 0, sizeof(*intfs));
+        *pp_mdns_ips = mdns_ips = malloc(sizeof(*mdns_ips));
+        if (mdns_ips == NULL) {
+                free(mdns_ips);
+                return (MDNS_ERROR);
+        }
+        memset(mdns_ips, 0, sizeof(*mdns_ips));
         *p_nb_intf = 1;
         return (0);
 }
-#endif // HAVE_IFADDRS
+#endif // HAVE_GETIFADDRS
 
 #else // _WIN32
 
@@ -151,9 +177,10 @@ mdns_is_interface_valuable(IP_ADAPTER_ADDRESSES *intf)
 }
 
 static size_t
-mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
+mdns_list_interfaces(multicast_if** pp_intfs, struct mdns_ip **pp_mdns_ips, size_t* p_nb_intf, int ai_family)
 {
         multicast_if* intfs;
+        struct mdns_ip *mdns_ips;
         IP_ADAPTER_ADDRESSES *res = NULL, *current;
         ULONG size;
         HRESULT hr;
@@ -176,8 +203,8 @@ mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
                 res = malloc( size );
                 if (res == NULL)
                         return (MDNS_ERROR);
-                hr = GetAdaptersAddresses(ai_family, GAA_FLAG_SKIP_UNICAST |
-                                                    GAA_FLAG_SKIP_ANYCAST | GAA_FLAG_SKIP_DNS_SERVER,
+                hr = GetAdaptersAddresses(ai_family, GAA_FLAG_SKIP_ANYCAST |
+                                                    GAA_FLAG_SKIP_DNS_SERVER,
                                                     NULL, res, &size);
         } while (hr == ERROR_BUFFER_OVERFLOW);
         if (hr != NO_ERROR) {
@@ -191,9 +218,20 @@ mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
                 ++nb_intf;
         }
         if (nb_intf == 0) {
-                *pp_intfs = malloc(sizeof(*intfs));
                 // Fallback to the default interface
+                *pp_intfs = malloc(sizeof(*intfs));
+                if (*pp_intfs == NULL) {
+                        free(res);
+                        return (MDNS_ERROR);
+                }
                 **pp_intfs = 0;
+                *pp_mdns_ips = malloc(sizeof(*mdns_ips));
+                if (*pp_mdns_ips == NULL) {
+                        free(*pp_intfs);
+                        free(res);
+                        return (MDNS_ERROR);
+                }
+                memset(mdns_ips, 0, sizeof(*mdns_ips));
                 *p_nb_intf = 1;
                 return (0);
         }
@@ -203,16 +241,45 @@ mdns_list_interfaces(multicast_if** pp_intfs, size_t* p_nb_intf, int ai_family)
                 free(res);
                 return (MDNS_ERROR);
         }
+        *pp_mdns_ips = mdns_ips = malloc(sizeof(*mdns_ips) * nb_intf);
+        if (mdns_ips == NULL) {
+                free(intfs);
+                free(res);
+                return (MDNS_ERROR);
+        }
         for (current = res; current != NULL; current = current->Next) {
                 if (!mdns_is_interface_valuable(current))
                         continue;
-                if (ai_family == AF_INET6)
-                        *intfs = htonl(current->Ipv6IfIndex);
-                else
+                if (ai_family == AF_INET6) {
+                        // For IPv6, The input value for setting IPV6_MULTICAST_IF is the
+                        // interface index of the desired outgoing interface in *host byte order*.
+                        // See https://docs.microsoft.com/en-us/windows/win32/winsock/ipproto-ipv6-socket-options
+                        *intfs = current->Ipv6IfIndex;
+                }
+                else {
+                        // For IPv4, The input value for setting IP_MULTICAST_IF is the
+                        // interface index of the desired outgoing interface in *network byte order*.
+                        // See https://docs.microsoft.com/en-us/windows/win32/winsock/ipproto-ip-socket-options
                         *intfs = htonl(current->IfIndex);
+                }
+                PIP_ADAPTER_UNICAST_ADDRESS_XP p_unicast = current->FirstUnicastAddress;
+                if( p_unicast )
+                {
+                        // Take the first unicast address (highest priority)
+                        if (ai_family == AF_INET) {
+                                struct sockaddr_in *addr_in = (struct sockaddr_in*) p_unicast->Address.lpSockaddr;
+                                memcpy(&(*mdns_ips).ipv4.addr, &addr_in->sin_addr, sizeof(struct in_addr));
+                        }
+                        else {
+                                struct sockaddr_in6 *addr_in = (struct sockaddr_in6*) p_unicast->Address.lpSockaddr;
+                                memcpy(&(*mdns_ips).ipv6.addr, &addr_in->sin6_addr, sizeof(struct in6_addr));
+                        }
+                }
+                ++mdns_ips;
                 ++intfs;
         }
         *p_nb_intf = nb_intf;
+        free(res);
         return (0);
 }
 #endif
@@ -223,6 +290,7 @@ mdns_resolve(struct mdns_ctx *ctx, const char *addr, unsigned short port)
         char buf[6];
         struct addrinfo hints, *res = NULL;
         multicast_if* ifaddrs = NULL;
+        struct mdns_ip *mdns_ips = NULL;
         size_t i;
         int status;
 
@@ -235,7 +303,7 @@ mdns_resolve(struct mdns_ctx *ctx, const char *addr, unsigned short port)
         if (errno != 0)
                 return (MDNS_LKPERR);
 
-        status = mdns_list_interfaces(&ifaddrs, &ctx->nb_conns, res->ai_family);
+        status = mdns_list_interfaces(&ifaddrs, &mdns_ips, &ctx->nb_conns, res->ai_family);
         if ( status < 0) {
                 freeaddrinfo(res);
                 return (status);
@@ -254,9 +322,11 @@ mdns_resolve(struct mdns_ctx *ctx, const char *addr, unsigned short port)
         for (i = 0; i < ctx->nb_conns; ++i ) {
                 ctx->conns[i].sock = INVALID_SOCKET;
                 ctx->conns[i].if_addr = ifaddrs[i];
-                ctx->conns[i].family = res->ai_family;
+                ctx->conns[i].mdns_ip = mdns_ips[i];
+                ctx->conns[i].mdns_ip.family = res->ai_family;
         }
         free(ifaddrs);
+        free(mdns_ips);
         freeaddrinfo(res);
         return (0);
 }
@@ -296,7 +366,7 @@ mdns_init(struct mdns_ctx **p_ctx, const char *addr, unsigned short port)
                 return mdns_destroy(ctx), (res);
 
         for (size_t i = 0; i < ctx->nb_conns; ++i ) {
-                if ((ctx->conns[i].sock = socket(ctx->conns[i].family, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET)
+                if ((ctx->conns[i].sock = socket(ctx->conns[i].mdns_ip.family, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET)
                         return mdns_destroy(ctx), (MDNS_NETERR);
                 if (setsockopt(ctx->conns[i].sock, SOL_SOCKET, SO_REUSEADDR, (const void *) &on_off, sizeof(on_off)) < 0)
                         return mdns_destroy(ctx), (MDNS_NETERR);
@@ -321,21 +391,21 @@ mdns_init(struct mdns_ctx **p_ctx, const char *addr, unsigned short port)
 
             if (os_mcast_join(ctx->conns[i].sock, &ctx->addr, ctx->conns[i].if_addr) < 0)
                     return mdns_destroy(ctx), (MDNS_NETERR);
-            if (setsockopt(ctx->conns[i].sock, ctx->conns[i].family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
-                           ctx->conns[i].family == AF_INET ? IP_MULTICAST_TTL : IPV6_MULTICAST_HOPS,
+            if (setsockopt(ctx->conns[i].sock, ctx->conns[i].mdns_ip.family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
+                           ctx->conns[i].mdns_ip.family == AF_INET ? IP_MULTICAST_TTL : IPV6_MULTICAST_HOPS,
                            (const void *) &ttl, sizeof(ttl)) < 0) {
                     return mdns_destroy(ctx), (MDNS_NETERR);
             }
 
-            if (setsockopt(ctx->conns[i].sock, ctx->conns[i].family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
+            if (setsockopt(ctx->conns[i].sock, ctx->conns[i].mdns_ip.family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
                            IP_MULTICAST_LOOP, (const void *) &loop, sizeof(loop)) < 0) {
                     return mdns_destroy(ctx), (MDNS_NETERR);
             }
 
-#if defined(HAVE_IFADDRS) || defined(_WIN32)
+#if defined(HAVE_GETIFADDRS) || defined(_WIN32)
             if (setsockopt(ctx->conns[i].sock,
-                           ctx->conns[i].family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
-                           ctx->conns[i].family == AF_INET ? IP_MULTICAST_IF : IPV6_MULTICAST_IF,
+                           ctx->conns[i].mdns_ip.family == AF_INET ? IPPROTO_IP : IPPROTO_IPV6,
+                           ctx->conns[i].mdns_ip.family == AF_INET ? IP_MULTICAST_IF : IPV6_MULTICAST_IF,
                            (const void*)&ctx->conns[i].if_addr, sizeof(ctx->conns[i].if_addr))) {
                     return mdns_destroy(ctx), (MDNS_NETERR);
             }
@@ -520,7 +590,7 @@ strrcmp(const char *s1, const char *s2)
 
 static int
 mdns_listen_probe_network(const struct mdns_ctx *ctx, const char *const names[],
-                          unsigned int nb_names, mdns_callback callback,
+                          unsigned int nb_names, mdns_listen_callback callback,
                           void *p_cookie)
 {
     struct mdns_hdr ahdr = {0};
@@ -569,7 +639,7 @@ mdns_listen_probe_network(const struct mdns_ctx *ctx, const char *const names[],
 int
 mdns_listen(const struct mdns_ctx *ctx, const char *const names[],
             unsigned int nb_names, enum rr_type type, unsigned int interval,
-            mdns_stop_func stop, mdns_callback callback, void *p_cookie)
+            mdns_stop_func stop, mdns_listen_callback callback, void *p_cookie)
 {
         if (ctx->nb_conns == 0)
                 return (MDNS_ERROR);
@@ -616,7 +686,7 @@ mdns_listen(const struct mdns_ctx *ctx, const char *const names[],
 
 int
 mdns_announce(struct mdns_ctx *ctx, const char *service, enum rr_type type,
-        mdns_callback callback, void *p_cookie)
+        mdns_announce_callback callback, void *p_cookie)
 {
         if (!callback)
                 return (MDNS_ERROR);
@@ -627,7 +697,7 @@ mdns_announce(struct mdns_ctx *ctx, const char *service, enum rr_type type,
 
         svc->name = strdup(service);
         svc->type = type;
-        svc->callback = callback;
+        svc->announce_callback = callback;
         svc->p_cookie = p_cookie;
         svc->next  = ctx->services;
 
@@ -669,7 +739,7 @@ mdns_serve(struct mdns_ctx *ctx, mdns_stop_func stop, void *p_cookie)
 
                         for (svc = ctx->services; svc; svc = svc->next) {
                                 if (!strrcmp(question->name, svc->name) && question->type == svc->type) {
-                                        svc->callback(svc->p_cookie, r, question);
+                                        svc->announce_callback(svc->p_cookie, r, &ctx->conns[i].mdns_ip, question);
                                         goto again;
                                 }
                         }
